#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

// globals for rendering
int fullscreen_mode=0;
int width=800, height=600;
int saved_width, saved_height;
int oldmousex=0, oldmousey=0;
enum {DONOTHING, ROTATE, ZOOM, MOVETARGET} mousemode=DONOTHING;
float camdist=6;
const float NEARCLIP_FACTOR=0.1, FARCLIP_FACTOR=10;
float nearclip, farclip;
float target[3]={0, 1, 0};
float view_target_timer=0;
int view_ground=1, view_buddha=1;
float old_time=0;
const float fovy=45;
float heading=M_PI/3, pitch=0;
float campos[3];
int file_count=0;

// globals for data
const char *buddhafilename=0;
int bnum_x, bnum_tri;
float *bx;
float *bn;
int *btri;
int frame=0;
const char *directory=0;
int n=0, allocn=0;
float *px=0;
float *pc=0;

// prototypes
void read_buddha();
void read_frame(int new_frame);
void handle_display();
void handle_reshape(int w, int h);
void handle_keypress(unsigned char key, int x, int y);
void handle_click(int button, int state, int x, int y);
void handle_drag(int x, int y);
void handle_idle();
void toggle_fullscreen();
void compute_camera_position();
void display_ground();
void display_buddha();
void display_particles();
void display_target();
void save_screen(const char *filename);
void init_lighting_and_material();

// definitions
int main(int argc, char *argv[])
{
   // initialize GLUT stuff 
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
   glutInitWindowSize(width,height);
   saved_width=width;
   saved_height=height;
   glutCreateWindow("hw2_viewer");
   glutDisplayFunc(handle_display);
   glutReshapeFunc(handle_reshape);
   glutKeyboardFunc(handle_keypress);
   glutMouseFunc(handle_click);
   glutMotionFunc(handle_drag);
   glutIdleFunc(handle_idle);

   // initialize OpenGL stuff
   glEnable(GL_DEPTH_TEST);
   glClearColor(0, 0, 0, 0);
   glClearDepth(1);
   glPixelStorei(GL_PACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   nearclip=NEARCLIP_FACTOR*camdist;
   farclip=FARCLIP_FACTOR*camdist;
   init_lighting_and_material();

   if(argc<3){
      fprintf(stderr, "Expecting <buddha_file> <particle_directory> as arguments\n");
      return 1;
   }
   buddhafilename=argv[1];
   directory=argv[2];

   // read in Buddha model
   read_buddha();

   // initialize first frames
   read_frame(0);

   // let's go
   old_time=clock()/(float)CLOCKS_PER_SEC;
   glutPostRedisplay();
   glutMainLoop();
   return 0;
}

void init_lighting_and_material()
{
   {
      GLfloat ambient[4] = {.3, .3, .3, 1};
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
   }
   {
      GLfloat color[4] = {.7, .7, .7, 1};
      glLightfv(GL_LIGHT0, GL_DIFFUSE, color);
      glLightfv(GL_LIGHT0, GL_SPECULAR, color);
      glEnable(GL_LIGHT0);
   }
   {
      GLfloat color[4] = {.4, .4, .4, 1};
      glLightfv(GL_LIGHT1, GL_DIFFUSE, color);
      glLightfv (GL_LIGHT1, GL_SPECULAR, color);
      glEnable (GL_LIGHT1);
   }
   {
      GLfloat color[4] = {.2, .2, .2, 1};
      glLightfv (GL_LIGHT2, GL_DIFFUSE, color);
      glLightfv (GL_LIGHT2, GL_SPECULAR, color);
      glEnable (GL_LIGHT2);
   }
   {
      GLfloat f_ambient[4] = {.5, .5, 0};
      GLfloat f_diffuse[4] = {.5, .3, .1};
      GLfloat f_specular[4] = {.5, 0, 0};
      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, f_ambient);
      glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, f_diffuse);
      glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, f_specular);
      glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 64);
   }
}

void read_buddha()
{
   FILE *fp;
   int i;

   fp=fopen(buddhafilename, "rt");
   fscanf(fp, "%d", &bnum_x);
   bx=(float*)malloc(3*bnum_x*sizeof(float));
   for(i=0; i<3*bnum_x; ++i)
      fscanf(fp, "%g", bx+i);
   fscanf(fp, "%d", &bnum_tri);
   btri=(int*)malloc(3*bnum_tri*sizeof(int));
   for(i=0; i<3*bnum_tri; ++i)
      fscanf(fp, "%d", btri+i);
   fclose(fp);
   // compute vertex normals
   bn=(float*)malloc(3*bnum_x*sizeof(float));
   for(i=0; i<bnum_tri; ++i){
      int a=btri[3*i], b=btri[3*i+1], c=btri[3*i+2];
      float normal[3], mag;
      normal[0]=(bx[3*b+1]-bx[3*a+1])*(bx[3*c+2]-bx[3*a+2])-(bx[3*b+2]-bx[3*a+2])*(bx[3*c+1]-bx[3*a+1]);
      normal[1]=(bx[3*b+2]-bx[3*a+2])*(bx[3*c+0]-bx[3*a+0])-(bx[3*b+0]-bx[3*a+0])*(bx[3*c+2]-bx[3*a+2]);
      normal[2]=(bx[3*b+0]-bx[3*a+0])*(bx[3*c+1]-bx[3*a+1])-(bx[3*b+1]-bx[3*a+1])*(bx[3*c+0]-bx[3*a+0]);
      mag=sqrt(normal[0]*normal[0]+normal[1]*normal[1]+normal[2]*normal[2]);
      normal[0]/=mag;
      normal[1]/=mag;
      normal[2]/=mag;
      bn[3*a]+=normal[0]; bn[3*b]+=normal[0]; bn[3*c]+=normal[0];
      bn[3*a+1]+=normal[1]; bn[3*b+1]+=normal[1]; bn[3*c+1]+=normal[1];
      bn[3*a+2]+=normal[2]; bn[3*b+2]+=normal[2]; bn[3*c+2]+=normal[2];
   }
   for(i=0; i<bnum_x; ++i){
      float mag=sqrt(bn[3*i+0]*bn[3*i+0]+bn[3*i+1]*bn[3*i+1]+bn[3*i+2]*bn[3*i+2]);
      bn[3*i]/=mag;
      bn[3*i+1]/=mag;
      bn[3*i+2]/=mag;
   }
}

#define LINE_SIZE 1000
void read_frame(int new_frame)
{
   char line[LINE_SIZE];
   char *file_name;
   FILE *fp;
   int i;

   if(new_frame<0) return;
   file_name=malloc(strlen(directory)+20);
   sprintf(file_name, "%s/particles%04d", directory, new_frame);
   fp=fopen(file_name, "rt");
   if(fp){
      int new_n;
      frame=new_frame;
      fscanf(fp, "%d", &new_n);
      if(px==0){
         allocn=new_n;
         px=(float*)malloc(3*allocn*sizeof(float));
         pc=(float*)malloc(3*allocn*sizeof(float));
      }else if(new_n>allocn){
         allocn+=new_n;
         px=(float*)realloc(px, 3*allocn*sizeof(float));
         pc=(float*)realloc(pc, 3*allocn*sizeof(float));
      }
      n=new_n;
      for(i=0; i<n; ++i){
         fgets(line, LINE_SIZE, fp);
         sscanf(line, "%f %f %f %f %f %f", px+3*i, px+3*i+1, px+3*i+2, pc+3*i, pc+3*i+1, pc+3*i+2);
      }
      fclose(fp);
   }
   free(file_name);
}

void handle_display()
{
   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

   glMatrixMode (GL_MODELVIEW);
   glLoadIdentity();
   gluLookAt(campos[0], campos[1], campos[2], target[0], target[1], target[2], 0, 1, 0);

   if (view_ground) display_ground();
   if (view_buddha) display_buddha();
   display_particles();
   if(view_target_timer>0) display_target();

   glutSwapBuffers();
}

void handle_reshape(int w, int h)
{
   width=w;
   height=h;
   glViewport(0, 0, (GLsizei)width, (GLsizei)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(fovy, width/(float)height, nearclip, farclip);
   glutPostRedisplay();
}

void handle_keypress(unsigned char key, int x, int y)
{
   switch(key){
   case '.':
      read_frame(frame+1);
      break;
   case ',':
      read_frame(frame-1);
      break;
   case 'b':
      view_buddha=!view_buddha;
      break;
   case 'f':
      toggle_fullscreen();
      break;
   case 'g':
      view_ground=!view_ground;
      break;
   case 'q':
      exit(0);
   case 's':
      {
         char file_name[30];
         sprintf(file_name, "hw1_%04d.ppm", file_count);
         save_screen(file_name);
         ++file_count;
      }
      break;
   default:
      return;
   }
   glutPostRedisplay();
}

void handle_click(int button, int state, int x, int y)
{
   int mods=glutGetModifiers();
   if(state==GLUT_UP) mousemode = DONOTHING;
   else if(mods&GLUT_ACTIVE_SHIFT){
      view_target_timer=1;
      mousemode=MOVETARGET;
   }else{
      if(button==GLUT_RIGHT_BUTTON || (button==GLUT_LEFT_BUTTON && mods&GLUT_ACTIVE_CTRL))
         mousemode=ZOOM;
      else if(button==GLUT_LEFT_BUTTON)
         mousemode=ROTATE;
   }
   oldmousex=x;
   oldmousey=y;
}

void handle_drag(int x, int y)
{
   switch(mousemode){
   case ROTATE:
      heading+=0.007*(oldmousex-x);
      if(heading<-M_PI) heading+=2*M_PI;
      else if(heading>M_PI) heading-=2*M_PI;
      pitch+=0.007*(oldmousey-y);
      if(pitch<-.48*M_PI) pitch=-.48*M_PI;
      else if(pitch>.48*M_PI) pitch=.48*M_PI;
      break;
   case ZOOM:
      camdist*=pow(1.01, y-oldmousey);
      nearclip=NEARCLIP_FACTOR*camdist;
      farclip=FARCLIP_FACTOR*camdist;
      break;
   case MOVETARGET:
      target[0]-=(0.002*camdist)*cos(heading)*(oldmousex-x);
      target[1]-=(0.002*camdist)*(oldmousey-y);
      target[2]-=(0.002*camdist)*sin(heading)*(oldmousex-x);
      break;
   case DONOTHING:
      break;
   }
   oldmousex=x;
   oldmousey=y;
   compute_camera_position();
   glutPostRedisplay();
}

void handle_idle()
{
   float new_time=clock()/(float)CLOCKS_PER_SEC;
   float delta_time=new_time-old_time;
   int target_turned_off=0;
   if(delta_time<0) delta_time=0; // this can happen due to roll-over
   old_time=new_time;
   if(view_target_timer>0){
      view_target_timer-=delta_time;
      if(view_target_timer<=0){
         target_turned_off=1;
         view_target_timer=0;
      }
   }
   if(target_turned_off)
      glutPostRedisplay();
}

void toggle_fullscreen()
{
   if(fullscreen_mode){
      fullscreen_mode=0;
      width=saved_width;
      height=saved_height;
      glutReshapeWindow(saved_width, saved_height);
   }else{
      fullscreen_mode=1;
      saved_width=width;
      saved_height=height;
      glutFullScreen();
   }
}

void compute_camera_position()
{
   float ch=cos(heading), sh=sin(heading), cp=cos(pitch), sp=sin(pitch);
   campos[0]=target[0]-camdist*sh*cp;
   campos[1]=target[1]-camdist*sp;
   campos[2]=target[2]-camdist*ch*cp;
}

void display_ground()
{
   glBegin(GL_QUADS);
   glColor3f(.2, .5, .5);
   glVertex3f(-3, 0, -3);
   glVertex3f(3, 0, -3);
   glVertex3f(3, 0, 3);
   glVertex3f(-3, 0, 3);
   glEnd();
}

void display_buddha()
{
   int i;
   glEnable (GL_LIGHTING);
   glColor3f(0.6, .5, 0);
   glBegin(GL_TRIANGLES);
   for(i=0; i<bnum_tri; ++i){
      glNormal3fv(bn+3*btri[3*i]);
      glVertex3fv(bx+3*btri[3*i]);
      glNormal3fv(bn+3*btri[3*i+1]);
      glVertex3fv(bx+3*btri[3*i+1]);
      glNormal3fv(bn+3*btri[3*i+2]);
      glVertex3fv(bx+3*btri[3*i+2]);
   }
   glEnd();
   glDisable (GL_LIGHTING);
}

void display_particles()
{
   int i;
   glBegin(GL_POINTS);
   for(i=0; i<n; ++i){
      glColor3fv(pc+3*i);
      glVertex3fv(px+3*i);
   }
   glEnd();
}

void display_target()
{
   float logcd=log(camdist)/log(20.);
   float smallsize=pow(20.0, floor(logcd-.5));
   glMatrixMode (GL_MODELVIEW);
   glPushMatrix();
   glTranslatef(target[0], target[1], target[2]);
   glColor3f(1, 1, 1);
   glutWireCube(smallsize);
   glutWireCube(20*smallsize);
   glPopMatrix();
}

void save_screen(const char *filename)
{
   FILE *fp;
   GLubyte *image_buffer;

   // get the pixels
   image_buffer=malloc(3*width*height*sizeof(GLubyte));
   glReadBuffer(GL_FRONT);
   glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, image_buffer);
  
   // write them to the file
   fp=fopen(filename, "wb");
   if(fp){
      int i;
      fprintf(fp, "P6\n%d %d 255\n", width, height);
      for(i=1; i<=height; ++i){
         fwrite(image_buffer+3*width*(height-i), 1, 3*width, fp);
      }
   }
   
   // clean up
   free(image_buffer);
}

